import os, sys
import pygame
import time
import random
import math
from pygame.locals import *

from pygameUtils import *


TILE_HEIGHT = 30
TILE_WIDTH = 30

class Character (pygame.sprite.Sprite):
    def __init__ (self, image, colorkey = -1):
        pygame.sprite.Sprite.__init__ (self)
        self.rect = pygame.Rect ((0,0) , (0,0))
        self.image = None
        self.location = (0,0)
        self.up = 1
        self.right = 2
        self.down = 3
        self.left = 4
        self.direction = 0
        self.lastDirection = 0
        self.nextDirection = 0
        self.originalImage, self.originalRect = load_image (image, colorkey)
        self.image = self.originalImage
        self.rect = self.originalRect
        self.locOffsetTop = self.originalImage.get_height()
        self.locOffsetLeft = self.originalImage.get_width()
        self.walkSpeed = 3
        self.map = None
        self.type = 'Character'
    def setMap (self, map):
        self.map = map
    def setSize (self, height, width):
        center = self.getCenter()
        height = int (height)
        width = int (width)
        self.image = pygame.transform.scale (self.originalImage, (height, width))
        self.rect = self.image.get_rect()
        self.setCenter (center [0], center [1])
    def getSize (self):
        return (self.rect.height, self.rect.width)
    def setLoc (self, top, left):
        self.location = (top, left)
        self.setAbsPosition (top * self.locOffsetTop, left * self.locOffsetLeft)
    def getRect (self):
        return self.rect
    def setAbsPosition (self, top, left):
        self.rect.top = top
        self.rect.left = left
    def getCenter (self):
        return self.rect.center
    def setCenter (self, top, left):
        self.rect.center = (top, left)
    def getSprites (self):
        pass
    def moveUp (self):
        # Set up cycling through the animations for walking down
        self.direction = self.up
    def moveDown (self):
        # Set up cycling through the animations for walking up
        self.direction = self.down
    def moveLeft (self):
        self.direction = self.left
    def moveRight (self):
        self.direction = self.right
    def getAbsPosition (self):
        return (self.rect.top, self.rect.left)
    def walk (self):
        if (self.direction == 0):
            return

        top, left = self.getAbsPosition()
        topMove = 0
        leftMove = 0
        if (self.direction == self.up):
            topMove = -1
        if (self.direction == self.down):
            topMove = 1
        if (self.direction == self.left):
            leftMove = -1
        if (self.direction == self.right):
            leftMove = 1

        arrived = True
        absTopDest, absLeftDest = self.getAbsDestination()
        #print "I'm at ", left
        #print "I'm going to ", absLeftDest
        if (topMove > 0):
            arrived = arrived and (top >= absTopDest)
        if (topMove < 0):
            arrived = arrived and (top <= absTopDest)
        if (leftMove > 0):
            arrived = arrived and (left >= absLeftDest)
        if (leftMove < 0):
            arrived = arrived and (left <= absLeftDest)
        if (arrived and self.direction != 0):
            self.direction = 0
            currTopLoc = self.location [0]
            currLeftLoc = self.location [1]
            self.setLoc ((currTopLoc + topMove),(currLeftLoc + leftMove))
            self.direction = self.nextDirection
            self.nextDirection = 0

        if (self.direction != 0):
            self.setAbsPosition (self.rect.top + (topMove * self.walkSpeed), self.rect.left + (leftMove * self.walkSpeed))
            
    def getAbsDestination (self):
        absTop, absLeft = self.getAbsPosition()
        locTop = self.location [0]
        locLeft = self.location [1]
        topMove = 0
        leftMove = 0
        if (self.direction == self.up):
            topMove = -1
        if (self.direction == self.down):
            topMove = 1
        if (self.direction == self.left):
            leftMove = -1
        if (self.direction == self.right):
            leftMove = 1
        destTop = locTop + topMove
        destLeft = locLeft + leftMove
        absTopDest = self.locOffsetTop * destTop
        absLeftDest = self.locOffsetLeft * destLeft
        return (absTopDest, absLeftDest)
    def update (self):
        self.walk()
        center = self.getCenter()
        pygame.sprite.Sprite.update (self)
    def setDirection (self, direct):
        if (self.direction == 0):
            currTop = self.location [0]
            currLeft = self.location [1]
            nextTop = currTop
            nextLeft = currLeft
            if (direct == self.up):
                nextTop = nextTop - 1
            elif (direct == self.right):
                nextLeft = nextLeft + 1
            elif (direct == self.left):
                nextLeft = nextLeft - 1
            elif (direct == self.down):
                nextTop = nextTop + 1
            if (self.map.getTile (nextTop, nextLeft).passable):
                self.direction = direct
                self.lastDirection = direct
                if (self.direction == self.up):
                    self.image = self.originalImage
                elif (self.direction == self.right):
                    self.image = pygame.transform.rotate (self.originalImage, 270)
                elif (self.direction == self.down):
                    self.image = pygame.transform.rotate (self.originalImage, 180)
                elif (self.direction == self.left):
                    self.image = pygame.transform.rotate (self.originalImage, 90)

                return True
            else:
                return False
        else:
            #self.nextDirection = direct
            pass

class Mummy (Character):
    def __init__ (self, image):
        Character.__init__ (self, image)
        self.type = 'Mummy'
        self.stunned = False
        self.sound = load_sound ('moan3.wav')
        self.setGuyLoc (0,0)
    def getSprites (self):
        return list()
    def walk (self):
        if (not self.stunned):
            Character.walk (self)
    def setGuyLoc (self, left, top):
        self.guyLoc = (left, top)
    def update (self):
        if (random.random() > 0.995):
            self.sound.stop()
            self.sound.play()
        if (self.direction == 0):
            moves = list()
            moves.append (self.up)
            moves.append (self.down)
            moves.append (self.right)
            moves.append (self.left)
            decided = False
            if (not decided and self.lastDirection != 0):
                if (random.random() * 1000 > 100):
                    decided = self.setDirection (self.lastDirection)
            if (not decided and self.guyLoc [0] < self.location [0]):
                decided = self.setDirection (self.up)
            if (not decided and self.guyLoc [0] > self.location [0]):
                decided = self.setDirection (self.down)
            if (not decided and self.guyLoc [1] < self.location [1]):
                decided = self.setDirection (self.left)
            if (not decided and self.guyLoc [1] > self.location [1]):
                decided = self.setDirection (self.right)
            while (not decided):
                choice = random.choice (moves)
                decided = self.setDirection (choice)
        Character.update (self)
    def stun (self):
        self.stunned = True
    def unstun (self):
        self.stunned = False

class Guy (Character):
    def __init__ (self, image):
        Character.__init__ (self, image)
        self.lightRing = LightRing ('lightRing.gif')
        self.setLoc (0,0)
        self.type = 'Guy'
        self.points = 0
        self.torch = 200
        self.torchShrink = 0
        self.lightRing.setSize (self.torch, self.torch)
    def moveUp (self):
        # Set up cycling through the animations for walking down
        self.direction = self.up
    def moveDown (self):
        # Set up cycling through the animations for walking up
        self.direction = self.down
    def moveLeft (self):
        self.direction = self.left
    def moveRight (self):
        self.direction = self.right
    def getSprites (self):
        return self.lightRing
    def __checkTorch (self):
        if (self.torch <= 60):
            return
        self.torchShrink = self.torchShrink + 1
        if (self.torchShrink > 6):
            self.torch = self.torch - 1
            self.torchShrink = 0
        if (self.torch % 5 == 0):
            self.refreshTorchSize()
    def refreshTorchSize (self):
        self.lightRing.setSize (self.torch, self.torch)
    def update (self):
        #self.moveDown()
        self.__checkTorch()
        center = self.getCenter()
        self.lightRing.setCenter (center [0], center [1])
        Character.update (self)

class LightRing (Character):
    def __init__ (self, image):
        Character.__init__ (self, image, (255,255,255))
    def update (self):
#        h = self.getSize() [0]
#        w = self.getSize() [1]
#        self.setSize (w + 1, h + 1)
        Character.update (self)
    def setSize (self, height, width):
        Character.setSize (self, height + 5, width + 5)

def __main__():
    stillPlaying = True
    random.seed()

    cracking = False
    crackSeq = 0
    
    pygame.init()
    screen = pygame.display.set_mode ((640, 480))

    whoosh = load_sound ('whoosh.wav')
    ugh = load_sound ('ugh2.wav')

    clippingOn = True
    background = pygame.Surface (screen.get_size())
    background = background.convert()
    background.fill ((0, 0, 0))
    guy = Guy ('guy.gif')
    mummy = Mummy ('mummy.gif')
    spritesList = list()
    lightsList = list()
    tilesList = list()
    enemiesList = list()
    theMap = Map (0)
    tilesList.extend (theMap.getSprites())
    enemiesList.append (mummy)
    spritesList.extend (theMap.getObjects())
    spritesList.append (guy)
    lightsList.append (guy.getSprites())
    spritesGroup = pygame.sprite.Group()
    lightsGroup = pygame.sprite.Group()
    tilesGroup = pygame.sprite.Group()
    enemiesGroup = pygame.sprite.Group()
    for x in spritesList:
        x.setMap (theMap)
        spritesGroup.add (x)
    for x in lightsList:
        x.setMap (theMap)
        lightsGroup.add (x)
    for x in tilesList:
        x.setMap (theMap)
        tilesGroup.add (x)
    for x in enemiesList:
        x.setMap (theMap)
        enemiesGroup.add (x)
    #allsprites = pygame.sprite.RenderUpdates (spritesGroup)
    allsprites = pygame.sprite.RenderUpdates (spritesGroup)
    alltiles = pygame.sprite.RenderUpdates (tilesGroup)
    alllights = pygame.sprite.RenderUpdates (lightsGroup)
    allenemies = pygame.sprite.RenderUpdates (enemiesGroup)
    dirtyRects = list()
    screen.blit (background, (0,0))
    pygame.display.flip()

    startLoc = theMap.getStart()
    guy.setLoc (startLoc [0], startLoc [1])
    mumStart = theMap.mummyStart
    mummy.setLoc (mumStart [0], mumStart [1])

    class Lightning:
        def __init__ (self):
            self.cracking = False
            self.crackSeq = 0
            self.flashing = False
            self.sound = load_sound ('thunder_1.wav')
        def checkFlash (self):
            if (not self.cracking):
                self.cracking = (random.random() > 0.995)
                if (self.cracking):
                    self.sound.play()
            if (self.cracking):
                self.crackSeq = self.crackSeq + 1
                seq = self.crackSeq
                if (seq == 7):
                    for x in (theMap.getEdges()):
                        x.flash()
                    self.flashing = True
                if (seq == 13):
                    for x in (theMap.getEdges()):
                        x.unflash()
                    self.flashing = False
                if (seq == 17):
                    for x in (theMap.getEdges()):
                        x.flash()
                    self.flashing = True
                if (seq == 22):
                    for x in (theMap.getEdges()):
                        x.unflash()
                    self.crackSeq = 0
                    self.cracking = False
                    self.flashing = False
            return self.flashing

    lightning = Lightning()

    def setupClipping():
#        return
        if not clippingOn:
            return
        sampRect = lightsList [0].getRect()
        clipRect = Rect((0,0,0,0))
        clipRect.height = sampRect.height
        clipRect.width = sampRect.width
        clipRect.center = sampRect.center
        clipCenter = clipRect.center
        clipRect.height = clipRect.height - int (sampRect.height * .5)
        clipRect.width = clipRect.width - int (sampRect.width * .5)
        clipRect.center = clipCenter
        screen.set_clip (clipRect)

    def checkDirection():
        keystate = pygame.key.get_pressed()
        if (keystate [K_DOWN]):
            guy.setDirection (guy.down)
        elif (keystate [K_UP]):
            guy.setDirection (guy.up)
        elif (keystate [K_RIGHT]):
            guy.setDirection (guy.right)
        elif (keystate [K_LEFT]):
            guy.setDirection (guy.left)

    def gotTreasure():
        whoosh.play()
        guy.points = guy.points + 100
        theMap.treasureCount = theMap.treasureCount - 1
        if (theMap.treasureCount < 1):
            showScreen ('Nice job!', 100)
            return True
        return False            

    def gotTorch():
        whoosh.play()
        if (guy.torch < 100):
            guy.torch = guy.torch + 200
        else:
            guy.torch = guy.torch + 100
        guy.refreshTorchSize()

    def guyDied():
        ugh.stop()
        ugh.play()
        guy.direction = 0
        guy.setLoc (theMap.start [0], theMap.start [1])
        showScreen ("Game over. Press ENTER.", 40)

    def checkCollisions():
        collisions = pygame.sprite.spritecollide (guy, spritesGroup, False)
        collisions.extend (pygame.sprite.spritecollide (guy, enemiesGroup, False))
        for x in collisions:
            if (x.type == 'Treasure'):
                if (gotTreasure()):
                    return True
                x.kill()
            if (x.type == 'Torch'):
                gotTorch()
                x.kill()
            if (x.type == 'Mummy'):
                guyDied()
                return True
        return False

    flashing = False
    #clippingOn = False

    def showScreen (text, size):
        return
        showing = True
        myFont = pygame.font.Font (None, size)
        picture = myFont.render (text, False, (255, 255, 255), (0,0,0))
        picture.convert()
        screen.set_clip()
        while (showing):
            screen.blit (picture, (0,0))
            pygame.display.flip()
            states = pygame.key.get_pressed()
            if (states [K_RETURN]):
                showing = False
            screen.set_clip()
            pygame.event.get()
        screen.fill ((0,0,0))

    showScreen ("Press ENTER to loot the mummy's tomb!", 40)
    
    while stillPlaying:
        #get input
        for event in pygame.event.get():
            if event.type == QUIT or \
               (event.type == KEYDOWN and event.key == K_ESCAPE):
                stillPlaying = False
        checkDirection()
        for foo in dirtyRects:
            screen.blit (background, foo, foo)
        if (not flashing):
            setupClipping()
        alltiles.update()
        allsprites.update()
        alllights.update()
        allenemies.update()
        dirtyRects = alltiles.draw (screen)
        dirtyRects.extend (allsprites.draw (screen))
        dirtyRects.extend (allenemies.draw (screen))
        if (not flashing):
            dirtyRects.extend (alllights.draw (screen))
        pygame.display.update (dirtyRects)
        pygame.display.flip()
        screen.set_clip()
        time.sleep (.005)

        if (checkCollisions()):
            stillPlaying = False
        wasFlashing = flashing
        flashing = lightning.checkFlash()
        if (flashing):
            mummy.stun()
        elif (wasFlashing and not flashing):
            mummy.unstun()

        dist = math.sqrt ( \
            math.pow (abs (guy.location [0] - mummy.location [0]), 2) + \
            math.pow (abs (guy.location [1] - mummy.location [1]), 2))
        mummy.sound.set_volume (1.0 / (dist))
        mummy.setGuyLoc (guy.location [0], guy.location [1])
        
    pygame.display.quit()

class Tile (Character):
    def __init__ (self, image, colorkey = -1, passable = True):
        self.passable = passable
        Character.__init__ (self, image, colorkey)
        self.type = 'Tile'

class Edge (Tile):
    def __init__ (self, image1, image2, colorkey1 = -1):
        Tile.__init__ (self, image1, colorkey1, False)
        self.image2, self.rect2 = load_image (image2, colorkey1)
        self.type = 'Edge'
    def flash (self):
        #self.rect2 = self.rect
        self.image = self.image2
    def unflash (self):
        self.image = self.originalImage

class Torch (Tile):
    def __init__ (self, image, colorkey = -1):
        Tile.__init__ (self, image, colorkey, True)
        self.type = 'Torch'

class Wall (Tile):
    def __init__ (self, image, colorkey = -1):
        Tile.__init__ (self, image, colorkey, False)
        self.type = 'Wall'

class Hall (Tile):
    def __init__ (self, image, colorkey = -1, passable = True):
        Tile.__init__ (self, image, colorkey, passable)
        self.type = 'Hall'

class Treasure (Tile):
    def __init__ (self, image, colorkey = -1):
        Tile.__init__ (self, image, colorkey, True)
        self.type = 'Treasure'

class Map:
    def __init__ (self, level):
        self.map = dict()
        self.objects = dict()
        self.edges = dict()
        self.level = level
        self.start = (0,0)
        self.mummyStart = (0,0)
        file = open ('data/level' + str (level) + '.lev', 'r')
        # First line is height
        read = file.readline()
        while (read.find ('#') != -1):
            read = file.readline()
        height = read
        # Second line is width
        read = file.readline()
        while (read.find ('#') != -1):
            read = file.readline()
        width = read
        # Subsequent lines are everything in the map
        h = 0
        w = 0
        numTreasures = 0
        for line in range (int (height)):
            input = file.readline()
            if (input.find ('#') == -1):
                for tile in input:
                    #print tile
                    newTile = None
                    newObject = None
                    if (tile == 'W'):
                        newTile = Wall ('wall.gif', None)
                    elif (tile == 'H'):
                        newTile = Hall ('hall.gif', None)
                    elif (tile == 'S'):
                        newTile = Hall ('hall.gif', None)
                        self.start = (h,w)
                    elif (tile == 'E'):
                        newTile = Edge ('edge1.gif','edge2.gif', None)
                        self.edges [(h,w)] = newTile
                    elif (tile == 'T'):
                        newTile = Hall ('hall.gif', None)
                        newObject = Treasure ('treasure.gif')
                        numTreasures = numTreasures + 1
                    elif (tile == 'R'):
                        newTile = Hall ('hall.gif', None)
                        newObject = Torch ('torch.gif')
                    elif (tile == 'M'):
                        newTile = Hall ('hall.gif', None)
                        self.mummyStart = (h,w)
                    if (newTile != None):
                        self.map [(h,w)] = newTile
                        newTile.setLoc (h, w)
                    if (newObject != None):
                        self.objects [(h,w)] = newObject
                        newObject.setLoc (h, w)
                    w = w + 1
                w = 0
                h = h + 1
        self.treasureCount = numTreasures
    def getTile (self, top, left):
        return self.map [(top, left)]
    def getEdges (self):
        sprites = list()
        for x in self.edges.values():
            if (x != None):
                sprites.append (x)
        return sprites
    def getObjects (self):
        sprites = list()
        for x in self.objects.values():
            if (x != None):
                sprites.append (x)
        return sprites
    def getStart (self):
        return self.start
    def getSprites (self):
        sprites = list()
        for x in self.map.values():
            if (x != None):
                sprites.append (x)
        return sprites

if __name__ == '__main__':
    __main__()
    
